import inspect
from contextlib import contextmanager
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Iterator, Optional, Tuple, Type

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest

from ninja_extra.controllers.response import ControllerResponse

from ...dependency_resolver import get_injector
from .context import RouteContext

if TYPE_CHECKING:
    from ...controllers import APIController, Route


class RouteFunctionContext:
    def __init__(
        self, controller_instance: "APIController", **view_func_kwargs: Any
    ) -> None:
        self.controller_instance = controller_instance
        self.view_func_kwargs = view_func_kwargs


class RouteFunction(object):
    def __init__(
        self, route: "Route", controller: Optional[Type["APIController"]] = None
    ):
        self.route = route
        self.has_request_param = False
        self.controller: Optional[Type["APIController"]] = controller
        self.as_view = wraps(route.view_func)(self.get_view_function())
        self._resolve_api_func_signature_(self.as_view)

    def __call__(self, request: HttpRequest, *args: Any, **kwargs: Any) -> Any:
        context = self.get_route_execution_context(request, *args, **kwargs)
        return self.as_view(*args, context=context, **kwargs)

    def _get_required_api_func_signature(self) -> Tuple:
        skip_parameters = ["self", "request"]
        sig_inspect = inspect.signature(self.route.view_func)
        sig_parameter = []
        for parameter in sig_inspect.parameters.values():
            if parameter.name not in skip_parameters:
                sig_parameter.append(parameter)
            elif parameter.name == "request":
                self.has_request_param = True
        return sig_inspect, sig_parameter

    def _resolve_api_func_signature_(self, context_func: Callable) -> Callable:
        # Override signature
        sig_inspect, sig_parameter = self._get_required_api_func_signature()
        sig_replaced = sig_inspect.replace(parameters=sig_parameter)
        context_func.__signature__ = sig_replaced  # type: ignore
        return context_func

    def get_view_function(self) -> Callable:
        def as_view(context: RouteContext, *args: Any, **kwargs: Any) -> Any:
            with self._prep_controller_route_execution(context, **kwargs) as ctx:
                ctx.controller_instance.check_permissions()
                result = self.route.view_func(
                    ctx.controller_instance, *args, **ctx.view_func_kwargs
                )
            return self._process_view_function_result(result)

        as_view.get_route_function = lambda: self  # type:ignore
        return as_view

    def _process_view_function_result(self, result: Any) -> Any:
        """
        This process any an returned value from view_func
        and creates an api response if result is ControllerResponseSchema
        """

        if result and isinstance(result, ControllerResponse):
            return result.status_code, result.convert_to_schema()
        return result

    def _get_controller_instance(self) -> "APIController":
        injector = get_injector()
        assert self.controller

        controller_instance: "APIController" = injector.create_object(self.controller)
        return controller_instance

    def get_route_execution_context(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> RouteContext:

        if not self.controller:
            raise ImproperlyConfigured("Controller object is required")

        init_kwargs = dict(
            permission_classes=self.route.permissions
            or self.controller.permission_classes,
            request=request,
            kwargs=kwargs,
            args=args,
        )
        context = RouteContext(**init_kwargs)
        return context

    @contextmanager
    def _prep_controller_route_execution(
        self, context: RouteContext, **kwargs: Any
    ) -> Iterator[RouteFunctionContext]:
        controller_instance = self._get_controller_instance()
        controller_instance.context = context

        api_func_kwargs = dict(**kwargs)
        if self.has_request_param:
            api_func_kwargs.update(request=context.request)
        try:
            yield RouteFunctionContext(
                controller_instance=controller_instance, **api_func_kwargs
            )
        except Exception as ex:
            raise ex
        finally:
            controller_instance.context = None

    def __str__(self) -> str:
        return self.route.route_params.path

    def __repr__(self) -> str:
        if not self.controller:
            return f"<RouteFunction, controller: No Controller Found, path: {self.__str__()}>"
        return f"<RouteFunction, controller: {self.controller.__name__}, path: {self.__str__()}>"


class AsyncRouteFunction(RouteFunction):
    def get_view_function(self) -> Callable:
        async def as_view(context: RouteContext, *args: Any, **kwargs: Any) -> Any:
            with self._prep_controller_route_execution(context, **kwargs) as ctx:
                ctx.controller_instance.check_permissions()
                result = await self.route.view_func(
                    ctx.controller_instance, *args, **ctx.view_func_kwargs
                )
            return self._process_view_function_result(result)

        as_view.get_route_function = lambda: self  # type:ignore
        return as_view

    def __repr__(self) -> str:
        if not self.controller:
            return f"<AsyncRouteFunction, controller: No Controller Found, path: {self.__str__()}>"
        return f"<AsyncRouteFunction, controller: {self.controller.__name__}, path: {self.__str__()}>"

    async def __call__(self, request: HttpRequest, *args: Any, **kwargs: Any) -> Any:
        context = self.get_route_execution_context(request, *args, **kwargs)
        return await self.as_view(*args, context=context, **kwargs)



from functools import wraps
from unittest.mock import patch

from ninja.signature import is_async


def mock_signal_call(signal: str, called: bool = True):
    def _wrap(func):
        if is_async(func):

            async def _wrapper(*args, **kwargs):
                with patch(f"ninja_extra.signals.{signal}.send") as mock_:
                    await func(*args, **kwargs)
                    assert mock_.called == called

        else:

            def _wrapper(*args, **kwargs):
                with patch(f"ninja_extra.signals.{signal}.send") as mock_:
                    func(*args, **kwargs)
                    assert mock_.called == called

        return wraps(func)(_wrapper)

    return _wrap


def mock_log_call(level: str, called: bool = True):
    def _wrap(func):
        if is_async(func):

            async def _wrapper(*args, **kwargs):
                with patch(
                    f"ninja_extra.logger.request_logger.{level.lower()}"
                ) as mock_:
                    await func(*args, **kwargs)
                    assert mock_.called == called

        else:

            def _wrapper(*args, **kwargs):
                with patch(
                    f"ninja_extra.logger.request_logger.{level.lower()}"
                ) as mock_:
                    func(*args, **kwargs)
                    assert mock_.called == called

        return wraps(func)(_wrapper)

    return _wrap



import time
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterator,
    List,
    Optional,
    Sequence,
    Union,
)

from django.http import HttpRequest
from django.http.response import HttpResponse, HttpResponseBase
from ninja.constants import NOT_SET
from ninja.operation import (
    AsyncOperation as NinjaAsyncOperation,
    Operation as NinjaOperation,
    PathView as NinjaPathView,
)
from ninja.signature import is_async

from ninja_extra.logger import request_logger
from ninja_extra.signals import route_context_finished, route_context_started

from .controllers.route.context import RouteContext

if TYPE_CHECKING:
    from .controllers import RouteFunction


class Operation(NinjaOperation):
    def _log_action(
        self,
        logger: Callable[..., Any],
        request: HttpRequest,
        duration: Optional[float] = None,
        ex: Optional[Exception] = None,
        **kwargs: Any,
    ) -> None:
        try:
            route_function: "RouteFunction" = (
                self.view_func.get_route_function()  # type:ignore
            )
            msg = (
                f'"{request.method.upper() if request.method else "METHOD NOT FOUND"} - '
                f'{route_function.__class__.__name__}[{self.view_func.__name__}] {request.path}" '
                f"{duration if duration else str(ex)}"
            )
            logger(msg, **kwargs)
        except (Exception,):
            pass

    def get_execution_context(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> RouteContext:
        route_function: "RouteFunction" = (
            self.view_func.get_route_function()  # type:ignore
        )

        if not route_function:
            raise Exception("Route Function is found missing")

        if route_function and not hasattr(
            route_function, "get_route_execution_context"
        ):
            raise Exception("Invalid Route Function")

        return route_function.get_route_execution_context(request, *args, **kwargs)

    @contextmanager
    def _prep_run(self, request: HttpRequest, **kw: Any) -> Iterator:
        try:
            start_time = time.time()
            values = self._get_values(request, kw)
            context = self.get_execution_context(request, **values)
            # send route_context_started signal
            route_context_started.send(RouteContext, route_context=context)

            yield values, context
            self._log_action(
                request_logger.info,
                request=request,
                duration=time.time() - start_time,
                extra=dict(request=request),
                exc_info=None,
            )
        except Exception as e:
            self._log_action(
                request_logger.error,
                request=request,
                ex=e,
                extra=dict(request=request),
                exc_info=None,
            )
            raise e
        finally:
            # send route_context_finished signal
            route_context_finished.send(RouteContext, route_context=None)

    def run(self, request: HttpRequest, **kw: Any) -> HttpResponseBase:
        error = super(Operation, self)._run_checks(request)
        if error:
            return error
        try:
            with self._prep_run(request, **kw) as ctx:
                values, context = ctx
                result = self.view_func(context=context, **values)
                _processed_results = self._result_to_response(request, result)
            return _processed_results
        except Exception as e:
            if isinstance(e, TypeError) and "required positional argument" in str(e):
                msg = "Did you fail to use functools.wraps() in a decorator?"
                msg = f"{e.args[0]}: {msg}" if e.args else msg
                e.args = (msg,) + e.args[1:]
            return self.api.on_exception(request, e)


class AsyncOperation(Operation, NinjaAsyncOperation):
    async def run(self, request: HttpRequest, **kw: Any) -> HttpResponseBase:  # type: ignore
        error = self._run_checks(request)
        if error:
            return error
        try:
            with self._prep_run(request, **kw) as ctx:
                values, context = ctx
                result = await self.view_func(context=context, **values)
                _processed_results = self._result_to_response(request, result)
            return _processed_results
        except Exception as e:
            return self.api.on_exception(request, e)


class PathView(NinjaPathView):
    async def _async_view(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # type: ignore
        return await super(PathView, self)._async_view(request, *args, **kwargs)

    def _sync_view(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:  # type: ignore
        return super(PathView, self)._sync_view(request, *args, **kwargs)

    def add_operation(
        self,
        path: str,
        methods: List[str],
        view_func: Callable,
        *,
        auth: Optional[Union[Sequence[Callable], Callable, object]] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
    ) -> Operation:
        if url_name:
            self.url_name = url_name

        operation_class = Operation
        if is_async(view_func):
            self.is_async = True
            operation_class = AsyncOperation

        operation = operation_class(
            path,
            methods,
            view_func,
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
        )

        self.operations.append(operation)
        return operation



import django
import pytest

from ninja_extra import APIController, route, router
from ninja_extra.testing import TestAsyncClient, TestClient

from .utils import mock_log_call, mock_signal_call


class TestOperation:
    @router("")
    class SomeTestController(APIController):
        @route.get("/example")
        def example(self):
            return {"message": "example"}

        @route.get("/example_exception")
        def example_exception(self):
            raise Exception()

    @mock_signal_call("route_context_started")
    @mock_signal_call("route_context_finished")
    @mock_log_call("info")
    def test_route_operation_execution_works(self):
        client = TestClient(self.SomeTestController)
        response = client.get(str(self.SomeTestController.example))
        assert response.json() == {"message": "example"}

    @mock_signal_call("route_context_started")
    @mock_signal_call("route_context_finished")
    @mock_log_call("error")
    def test_route_operation_execution_should_log_execution(self):
        client = TestClient(self.SomeTestController)
        with pytest.raises(Exception):
            client.get(str(self.SomeTestController.example_exception))


@pytest.mark.skipif(django.VERSION < (3, 1), reason="requires django 3.1 or higher")
@pytest.mark.asyncio
class TestAsyncOperations:
    if not django.VERSION < (3, 1):

        @router("")
        class SomeTestController(APIController):
            @route.get("/example")
            async def example(self):
                return {"message": "example"}

            @route.get("/example_exception")
            async def example_exception(self):
                raise Exception()

        @mock_signal_call("route_context_started")
        @mock_signal_call("route_context_finished")
        @mock_log_call("info")
        async def test_async_route_operation_execution_works(self):
            client = TestAsyncClient(self.SomeTestController)
            response = await client.get(str(self.SomeTestController.example))
            assert response.json() == {"message": "example"}

        @mock_signal_call("route_context_started")
        @mock_signal_call("route_context_finished")
        @mock_log_call("error")
        async def test_async_route_operation_execution_should_log_execution(self):
            client = TestAsyncClient(self.SomeTestController)
            with pytest.raises(Exception):
                await client.get(str(self.SomeTestController.example_exception))



from urllib import parse

from django.utils.encoding import force_str


def replace_query_param(url: str, key: str, val: int) -> str:
    """
    Given a URL and a key/val pair, set or replace an item in the query
    parameters of the URL, and return the new URL.
    """
    (scheme, netloc, path, query, fragment) = parse.urlsplit(force_str(url))
    query_dict = parse.parse_qs(query, keep_blank_values=True)
    query_dict[force_str(key)] = [force_str("{}".format(val))]
    query = parse.urlencode(sorted(list(query_dict.items())), doseq=True)
    return parse.urlunsplit((scheme, netloc, path, query, fragment))


def remove_query_param(url: str, key: str) -> str:
    """
    Given a URL and a key/val pair, remove an item in the query
    parameters of the URL, and return the new URL.
    """
    (scheme, netloc, path, query, fragment) = parse.urlsplit(force_str(url))
    query_dict = parse.parse_qs(query, keep_blank_values=True)
    query_dict.pop(key, None)
    query = parse.urlencode(sorted(list(query_dict.items())), doseq=True)
    return parse.urlunsplit((scheme, netloc, path, query, fragment))



from typing import Any, List

from django.conf import settings as django_settings
from django.test.signals import setting_changed
from ninja import Schema
from pydantic import Field, root_validator, validator

from ninja_extra.lazy import LazyStrImport


class UserDefinedSettingsMapper:
    def __init__(self, data: dict) -> None:
        self.__dict__ = data


NinjaExtra_SETTINGS_DEFAULTS = dict(
    INJECTOR_MODULES=[],
    PAGINATION_CLASS="ninja_extra.pagination.LimitOffsetPagination",
)

USER_SETTINGS = UserDefinedSettingsMapper(
    getattr(django_settings, "NINJA_EXTRA", NinjaExtra_SETTINGS_DEFAULTS)
)


class NinjaExtraSettings(Schema):
    class Config:
        orm_mode = True
        validate_assignment = True

    PAGINATION_CLASS: Any = Field(
        "ninja_extra.pagination.PageNumberPaginationExtra",
    )
    PAGINATION_PER_PAGE: int = Field(100)
    INJECTOR_MODULES: List[Any] = []

    @validator("INJECTOR_MODULES", pre=True)
    def pre_injector_module_validate(cls, value: Any) -> Any:
        if not isinstance(value, list):
            raise ValueError("Invalid data type")
        return value

    @validator("PAGINATION_CLASS", pre=True)
    def pre_pagination_class_validate(cls, value: Any) -> Any:
        if isinstance(value, list):
            raise ValueError("Invalid data type")
        return value

    @root_validator
    def validate_ninja_extra_settings(cls, values: Any) -> Any:
        for item in NinjaExtra_SETTINGS_DEFAULTS.keys():
            if (
                isinstance(values[item], (tuple, list))
                and values[item]
                and isinstance(values[item][0], str)
            ):
                values[item] = [LazyStrImport(str(klass)) for klass in values[item]]
            if isinstance(values[item], str):
                values[item] = LazyStrImport(values[item])
        return values


# convert to lazy object
settings = NinjaExtraSettings.from_orm(USER_SETTINGS)


def reload_settings(*args: Any, **kwargs: Any) -> None:
    global settings

    setting, value = kwargs["setting"], kwargs["value"]

    if setting == "NINJA_EXTRA":
        settings = NinjaExtraSettings.from_orm(UserDefinedSettingsMapper(value))


setting_changed.connect(reload_settings)



"""
Copied from DRF
Provides a set of pluggable permission policies.
"""
from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Tuple, Type, TypeVar

from django.http import HttpRequest
from ninja.types import DictStrAny

if TYPE_CHECKING:
    from ninja_extra.controllers.base import APIController

SAFE_METHODS = ("GET", "HEAD", "OPTIONS")

T = TypeVar("T")


class OperationHolderMixin:
    def __and__(self, other: Type["BasePermission"]) -> "OperandHolder[AND]":
        return OperandHolder(AND, self, other)  # type: ignore

    def __or__(self, other: Type["BasePermission"]) -> "OperandHolder[OR]":
        return OperandHolder(OR, self, other)  # type: ignore

    def __rand__(self, other: Type["BasePermission"]) -> "OperandHolder[AND]":
        return OperandHolder(AND, other, self)  # type: ignore

    def __ror__(self, other: Type["BasePermission"]) -> "OperandHolder[OR]":
        return OperandHolder(OR, other, self)  # type: ignore

    def __invert__(self) -> "SingleOperandHolder[NOT]":
        return SingleOperandHolder(NOT, self)  # type: ignore


class BasePermissionMetaclass(OperationHolderMixin, ABCMeta):
    pass


class BasePermission(ABC, metaclass=BasePermissionMetaclass):
    """
    A base class from which all permission classes should inherit.
    """

    message: Any = None

    @abstractmethod
    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True

    def has_object_permission(
        self, request: HttpRequest, controller: "APIController", obj: Any
    ) -> bool:
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True


class SingleOperandHolder(OperationHolderMixin, Generic[T]):
    def __init__(
        self, operator_class: Type[BasePermission], op1_class: Type[BasePermission]
    ) -> None:
        super().__init__()
        self.operator_class = operator_class
        self.op1_class = op1_class

    def __call__(self, *args: Tuple[Any], **kwargs: DictStrAny) -> BasePermission:
        op1 = self.op1_class()
        return self.operator_class(op1)  # type: ignore


class OperandHolder(OperationHolderMixin, Generic[T]):
    def __init__(
        self,
        operator_class: Type["BasePermission"],
        op1_class: Type["BasePermission"],
        op2_class: Type["BasePermission"],
    ) -> None:
        self.operator_class = operator_class
        self.op1_class = op1_class
        self.op2_class = op2_class

    def __call__(self, *args: Tuple[Any], **kwargs: DictStrAny) -> BasePermission:
        op1 = self.op1_class()
        op2 = self.op2_class()
        return self.operator_class(op1, op2)  # type: ignore


class AND(BasePermission):
    def __init__(self, op1: "BasePermission", op2: "BasePermission") -> None:
        self.op1 = op1
        self.op2 = op2

    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        return self.op1.has_permission(request, controller) and self.op2.has_permission(
            request, controller
        )

    def has_object_permission(
        self, request: HttpRequest, controller: "APIController", obj: Any
    ) -> bool:
        return self.op1.has_object_permission(
            request, controller, obj
        ) and self.op2.has_object_permission(request, controller, obj)


class OR(BasePermission):
    def __init__(self, op1: "BasePermission", op2: "BasePermission") -> None:
        self.op1 = op1
        self.op2 = op2

    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        return self.op1.has_permission(request, controller) or self.op2.has_permission(
            request, controller
        )

    def has_object_permission(
        self, request: HttpRequest, controller: "APIController", obj: Any
    ) -> bool:
        return self.op1.has_object_permission(
            request, controller, obj
        ) or self.op2.has_object_permission(request, controller, obj)


class NOT(BasePermission):
    def __init__(self, op1: "BasePermission") -> None:
        self.op1 = op1

    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        return not self.op1.has_permission(request, controller)

    def has_object_permission(
        self, request: HttpRequest, controller: "APIController", obj: Any
    ) -> bool:
        return not self.op1.has_object_permission(request, controller, obj)



from typing import Dict, Optional, Union

from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from ninja.errors import HttpError

from ninja_extra import status


class APIException(HttpError):
    """
    Subclasses should provide `.status_code` and `.message` properties.
    """

    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    message = _("A server error occurred.")

    def __init__(
        self,
        message: Union[str, Dict[str, str], None] = None,
        status_code: Optional[int] = None,
    ) -> None:
        self.message = message or self.message  # type: ignore
        self.status_code = status_code or self.status_code
        super().__init__(status_code=self.status_code, message=self.message)

    def __str__(self) -> str:
        return self.message


class AuthenticationFailed(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    message = _("Incorrect authentication credentials.")


class NotAuthenticated(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    message = _("Authentication credentials were not provided.")


class PermissionDenied(APIException):
    status_code = status.HTTP_403_FORBIDDEN
    message = _("You do not have permission to perform this action.")


class NotFound(APIException):
    status_code = status.HTTP_404_NOT_FOUND
    message = _("Not found.")


class MethodNotAllowed(APIException):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED
    default_detail = _('Method "{method}" not allowed.')

    def __init__(
        self,
        method: str,
        detail: Optional[str] = None,
        status_code: Optional[int] = None,
    ) -> None:
        if detail is None:
            detail = force_str(self.default_detail).format(method=method)
        super().__init__(status_code=status_code, message=detail)



from typing import List, Type, Union

from ninja_extra.permissions.base import (
    BasePermission,
    OperandHolder,
    SingleOperandHolder,
)

PermissionType = Union[
    List[Type[BasePermission]], List[OperandHolder], List[SingleOperandHolder]
]



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)

from django.urls import URLPattern, path as django_path
from ninja.constants import NOT_SET
from ninja.types import DictStrAny
from ninja.utils import normalize_path

from ninja_extra.controllers.route.route_functions import RouteFunction
from ninja_extra.permissions.common import AllowAny

if TYPE_CHECKING:
    from ninja_extra import NinjaExtraAPI
    from ninja_extra.controllers.base import APIController
    from ninja_extra.types import PermissionType


class ControllerBorg:
    _shared_state_: Dict[str, Dict[str, Type["APIController"]]] = dict(
        controllers=dict()
    )

    def __init__(self) -> None:
        self.__dict__ = self._shared_state_

    def add_controller(self, controller: Type["APIController"]) -> None:
        if controller.auto_import:
            self._shared_state_["controllers"].update({str(controller): controller})

    def remove_controller(
        self, controller: Type["APIController"]
    ) -> Optional[Type["APIController"]]:
        if str(controller) in self._shared_state_["controllers"]:
            return self._shared_state_["controllers"].pop(str(controller))
        return None

    def clear_controller(self) -> None:
        self._shared_state_["controllers"] = dict()

    @classmethod
    def get_controllers(cls) -> Dict[str, Type["APIController"]]:
        return cls._shared_state_.get("controllers", dict())


class ControllerRegistry(ControllerBorg):
    def __init__(self) -> None:
        ControllerBorg.__init__(self)


class ControllerRouter:
    _tags: Optional[List[str]] = None
    _controller: Type["APIController"]

    def __init__(
        self,
        prefix: str,
        *,
        auth: Any = NOT_SET,
        tags: Optional[List[str]] = None,
        permissions: Optional["PermissionType"] = None,
        controller: Optional[Type["APIController"]] = None,
    ) -> None:
        self.prefix = prefix
        self.auth = auth
        self.tags = tags
        self.permission_classes = permissions or [AllowAny]

        if controller:
            self._controller = controller
            self(controller)

    @property
    def controller(self) -> Optional[Type["APIController"]]:
        return self._controller

    @property
    def tags(self) -> Optional[List[str]]:
        return self._tags

    @tags.setter
    def tags(self, value: Union[str, List[str], None]) -> None:
        tag: Optional[List[str]] = cast(Optional[List[str]], value)
        if tag and isinstance(value, str):
            tag = [value]
        self._tags = tag

    def __call__(self, controller: Type["APIController"]) -> Type["APIController"]:
        self._controller = controller
        controller.permission_classes = self.permission_classes  # type:ignore
        controller._router = self
        self.tags = self.tags or controller.tags
        ControllerRegistry().add_controller(controller)
        return controller

    @classmethod
    def _get_class_route_functions(
        cls, controller: "APIController"
    ) -> Iterator[RouteFunction]:
        for method in controller.__dict__.values():
            if isinstance(method, RouteFunction):
                yield method

    @property
    def path_operations(self) -> DictStrAny:
        return self._controller.get_path_operations()

    def set_api_instance(self, api: "NinjaExtraAPI") -> None:
        self._controller.api = api
        for path_view in self.path_operations.values():
            path_view.set_api_instance(api, self)

    def build_routers(self) -> List[Tuple[str, "ControllerRouter"]]:
        return [(self.prefix, self)]

    def urls_paths(self, prefix: str) -> Iterator[URLPattern]:
        for path, path_view in self.path_operations.items():
            path = path.replace("{", "<").replace("}", ">")
            route = "/".join([i for i in (prefix, path) if i])
            # to skip lot of checks we simply treat double slash as a mistake:
            route = normalize_path(route)
            route = route.lstrip("/")

            yield django_path(
                route, path_view.get_view(), name=cast(str, path_view.url_name)
            )

    def __repr__(self) -> str:
        return f"<controller - {self._controller.__name__}>"

    def __str__(self) -> str:
        return f"{self._controller.__name__}"


router = ControllerRouter



import inspect
import logging
import sys
from collections import OrderedDict
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Optional, Type, Union, cast

from django.core.paginator import InvalidPage, Page, Paginator
from django.db.models import QuerySet
from django.http import HttpRequest
from ninja import Schema
from ninja.constants import NOT_SET
from ninja.pagination import LimitOffsetPagination, PageNumberPagination, PaginationBase
from ninja.signature import has_kwargs
from ninja.types import DictStrAny
from pydantic import Field

from ninja_extra.conf import settings
from ninja_extra.exceptions import NotFound
from ninja_extra.schemas import PaginatedResponseSchema, get_paginated_response_schema
from ninja_extra.urls import remove_query_param, replace_query_param

logger = logging.getLogger()

if TYPE_CHECKING:
    from .controllers import APIController

__all__ = [
    "PageNumberPagination",
    "PageNumberPaginationExtra",
    "PaginationBase",
    "LimitOffsetPagination",
    "paginate",
]


def _positive_int(
    integer_string: Union[str, int], strict: bool = False, cutoff: Optional[int] = None
) -> int:
    """
    Cast a string to a strictly positive integer.
    """
    ret = int(integer_string)
    if ret < 0 or (ret == 0 and strict):
        raise ValueError()
    if cutoff:
        return min(ret, cutoff)
    return ret


class PageNumberPaginationExtra(PaginationBase):
    class Input(Schema):
        page: int = Field(1, gt=0)
        page_size: int = Field(100, lt=200)

    page_query_param = "page"
    page_size_query_param = "page_size"

    max_page_size = 200
    paginator_class = Paginator

    def __init__(
        self,
        page_size: int = settings.PAGINATION_PER_PAGE,
        max_page_size: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.page_size = page_size
        self.max_page_size = max_page_size or 200
        self.Input = self.create_input()  # type:ignore

    def create_input(self) -> Type[Input]:
        class DynamicInput(PageNumberPaginationExtra.Input):
            page: int = Field(1, gt=0)
            page_size: int = Field(self.page_size, lt=self.max_page_size)

        return DynamicInput

    def paginate_queryset(
        self, items: QuerySet, request: HttpRequest, **params: Any
    ) -> Any:

        pagination_input = cast(PageNumberPaginationExtra.Input, params["pagination"])
        page_size = self.get_page_size(pagination_input.page_size)
        current_page_number = pagination_input.page
        paginator = self.paginator_class(items, page_size)
        try:
            url = request.build_absolute_uri()
            page: Page = paginator.page(current_page_number)
            return self.get_paginated_response(base_url=url, page=page)
        except InvalidPage as exc:
            msg = "Invalid page. {page_number} {message}".format(
                page_number=current_page_number, message=str(exc)
            )
            raise NotFound(msg)

    def get_paginated_response(self, *, base_url: str, page: Page) -> DictStrAny:
        return OrderedDict(
            [
                ("count", page.paginator.count),
                ("next", self.get_next_link(base_url, page=page)),
                ("previous", self.get_previous_link(base_url, page=page)),
                ("results", list(page)),
            ]
        )

    @classmethod
    def get_response_schema(
        cls, response_schema: Union[Schema, Type[Schema]]
    ) -> Type[Schema]:
        if sys.version_info >= (3, 8):
            return PaginatedResponseSchema[response_schema]
        return get_paginated_response_schema(response_schema)

    def get_next_link(self, url: str, page: Page) -> Optional[str]:
        if not page.has_next():
            return None
        page_number = page.next_page_number()
        return replace_query_param(url, self.page_query_param, page_number)

    def get_previous_link(self, url: str, page: Page) -> Optional[str]:
        if not page.has_previous():
            return None
        page_number = page.previous_page_number()
        if page_number == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, page_number)

    def get_page_size(self, page_size: int) -> int:
        if page_size:
            try:
                return _positive_int(page_size, strict=True, cutoff=self.max_page_size)
            except (KeyError, ValueError):
                pass

        return self.page_size


def paginate(func_or_pgn_class: Any = NOT_SET, **paginator_params: Any) -> Callable:
    isfunction = inspect.isfunction(func_or_pgn_class)
    isnotset = func_or_pgn_class == NOT_SET

    pagination_class: Type[PaginationBase] = settings.PAGINATION_CLASS

    if isfunction:
        return _inject_pagination(func_or_pgn_class, pagination_class)

    if not isnotset:
        pagination_class = func_or_pgn_class

    def wrapper(func: Callable) -> Any:
        return _inject_pagination(func, pagination_class, **paginator_params)

    return wrapper


def _inject_pagination(
    func: Callable,
    paginator_class: Type[PaginationBase],
    **paginator_params: Any,
) -> Callable:
    func_modified = cast(Any, func)
    func_modified.has_kwargs = True
    if not has_kwargs(func_modified):
        func_modified.has_kwargs = False
        logger.debug(
            f"function {func_modified.__name__} should have **kwargs if you want to use pagination parameters"
        )

    paginator: PaginationBase = paginator_class(**paginator_params)
    paginator_kwargs_name = "pagination"

    @wraps(func_modified)
    def view_with_pagination(controller: "APIController", *args: Any, **kw: Any) -> Any:
        func_kwargs = dict(kw)
        if not func_modified.has_kwargs:
            func_kwargs.pop(paginator_kwargs_name)

        items = func(controller, *args, **func_kwargs)
        assert (
            controller.context and controller.context.request
        ), "Request object is None"
        return paginator.paginate_queryset(items, controller.context.request, **kw)

    view_with_pagination._ninja_contribute_args = [  # type: ignore
        (
            paginator_kwargs_name,
            paginator.Input,
            paginator.InputSource,
        ),
    ]

    return view_with_pagination



from .base import SAFE_METHODS, BasePermission
from .common import AllowAny, IsAdminUser, IsAuthenticated, IsAuthenticatedOrReadOnly

__all__ = [
    "BasePermission",
    "SAFE_METHODS",
    "AllowAny",
    "IsAuthenticated",
    "IsAdminUser",
    "IsAuthenticatedOrReadOnly",
]



import inspect
import uuid
from typing import Any, List, Optional, Type, Union, cast

from ninja.constants import NOT_SET
from ninja.signature import is_async
from ninja.types import TCallable

from ninja_extra.controllers.response import ControllerResponse
from ninja_extra.permissions import BasePermission
from ninja_extra.schemas import RouteParameter

from .route_functions import AsyncRouteFunction, RouteFunction

POST = "POST"
PUT = "PUT"
PATCH = "PATCH"
DELETE = "DELETE"
GET = "GET"
ROUTE_METHODS = [POST, PUT, PATCH, DELETE, GET]


class RouteInvalidParameterException(Exception):
    pass


class Route(object):
    permissions: Optional[Optional[List[Type[BasePermission]]]] = None

    def __init__(
        self,
        path: str,
        methods: List[str],
        *,
        auth: Any = NOT_SET,
        response: Union[Any, List[Any]] = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        permissions: Optional[List[Type[BasePermission]]] = None,
    ) -> None:

        if not isinstance(methods, list):
            raise RouteInvalidParameterException("methods must be a list")

        methods = list(map(lambda m: m.upper(), methods))
        not_valid_methods = list(set(methods) - set(ROUTE_METHODS))
        if not_valid_methods:
            raise RouteInvalidParameterException(
                f"Method {','.join(not_valid_methods)} not allowed"
            )

        _response = response
        if (
            inspect.isclass(response)
            and issubclass(response, ControllerResponse)  # type:ignore
        ) or isinstance(response, ControllerResponse):
            response = cast(ControllerResponse, response)
            _response = {response.status_code: response.get_schema()}
        elif isinstance(response, list):
            _response_computed = dict()
            for item in response:
                if (
                    inspect.isclass(item) and issubclass(item, ControllerResponse)
                ) or isinstance(item, ControllerResponse):
                    _response_computed.update({item.status_code: item.get_schema()})
                elif isinstance(item, dict):
                    _response_computed.update(item)
            if not _response_computed:
                raise RouteInvalidParameterException(
                    f"Invalid response configuration: {response}"
                )
            _response = _response_computed

        ninja_route_params = RouteParameter(
            path=path,
            methods=methods,
            auth=auth,
            response=_response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
        )
        self.route_params = ninja_route_params
        self.is_async = False
        self.permissions = permissions
        self.route_function_class = RouteFunction

    def __call__(self, view_func: TCallable) -> RouteFunction:
        if is_async(view_func):
            self.route_function_class = AsyncRouteFunction

        self.view_func = view_func
        self.route_params.operation_id = (
            self.route_params.operation_id
            or f"{str(uuid.uuid4())[:8]}_controller_{self.view_func.__name__}"
        )
        return self.route_function_class(route=self)

    @classmethod
    def get(
        cls,
        path: str,
        *,
        auth: Any = NOT_SET,
        response: Union[Any, List[Any]] = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        permissions: Optional[List[Type[BasePermission]]] = None,
    ) -> "Route":
        return Route(
            path,
            [GET],
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            permissions=permissions,
        )

    @classmethod
    def post(
        cls,
        path: str,
        *,
        auth: Any = NOT_SET,
        response: Union[Any, List[Any]] = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        permissions: Optional[List[Type[BasePermission]]] = None,
    ) -> "Route":
        return Route(
            path,
            [POST],
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            permissions=permissions,
        )

    @classmethod
    def delete(
        cls,
        path: str,
        *,
        auth: Any = NOT_SET,
        response: Union[Any, List[Any]] = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        permissions: Optional[List[Type[BasePermission]]] = None,
    ) -> "Route":
        return Route(
            path,
            [DELETE],
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            permissions=permissions,
        )

    @classmethod
    def patch(
        cls,
        path: str,
        *,
        auth: Any = NOT_SET,
        response: Union[Any, List[Any]] = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        permissions: Optional[List[Type[BasePermission]]] = None,
    ) -> "Route":
        return Route(
            path,
            [PATCH],
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            permissions=permissions,
        )

    @classmethod
    def put(
        cls,
        path: str,
        *,
        auth: Any = NOT_SET,
        response: Union[Any, List[Any]] = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        permissions: Optional[List[Type[BasePermission]]] = None,
    ) -> "Route":
        return Route(
            path,
            [PUT],
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            permissions=permissions,
        )

    @classmethod
    def generic(
        cls,
        path: str,
        *,
        methods: List[str],
        auth: Any = NOT_SET,
        response: Union[Any, List[Any]] = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        permissions: Optional[List[Type[BasePermission]]] = None,
    ) -> "Route":
        return Route(
            path,
            methods,
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            permissions=permissions,
        )


route = Route



from ninja import Schema

from ninja_extra import APIController, NinjaExtraAPI, route, router
from ninja_extra.pagination import (
    PageNumberPagination,
    PageNumberPaginationExtra,
    PaginationBase,
    paginate,
)
from ninja_extra.testing import TestClient

ITEMS = list(range(100))


class CustomPagination(PaginationBase):
    # only offset param, defaults to 5 per page
    class Input(Schema):
        skip: int

    def paginate_queryset(self, items, request, **params):
        skip = params["pagination"].skip
        return items[skip : skip + 5]


@router("")
class SomeAPIController(APIController):
    @route.get("/items_1")
    @paginate  # WITHOUT brackets (should use default pagination)
    def items_1(self, **kwargs):
        return ITEMS

    @route.get("/items_2")
    @paginate()  # with brackets (should use default pagination)
    def items_2(self, someparam: int = 0, **kwargs):
        # also having custom param `someparam` - that should not be lost
        return ITEMS

    @route.get("/items_3")
    @paginate(CustomPagination)
    def items_3(self, **kwargs):
        return ITEMS

    @route.get("/items_4")
    @paginate(PageNumberPaginationExtra, page_size=10)
    def items_4(self, **kwargs):
        return ITEMS

    @route.get("/items_5")
    @paginate(PageNumberPagination, page_size=10)
    def items_5_without_kwargs(self):
        return ITEMS


api = NinjaExtraAPI()
api.register_controllers(SomeAPIController)

client = TestClient(SomeAPIController)


class TestPagination:
    def test_case1(self):
        response = client.get("/items_1?limit=10").json()
        assert response == ITEMS[:10]

        schema = api.get_openapi_schema()["paths"]["/api/items_1"]["get"]
        # print(schema)
        assert schema["parameters"] == [
            {
                "in": "query",
                "name": "limit",
                "schema": {
                    "title": "Limit",
                    "default": 100,
                    "exclusiveMinimum": 0,
                    "type": "integer",
                },
                "required": False,
            },
            {
                "in": "query",
                "name": "offset",
                "schema": {
                    "title": "Offset",
                    "default": 0,
                    "exclusiveMinimum": -1,
                    "type": "integer",
                },
                "required": False,
            },
        ]

    def test_case2(self):
        response = client.get("/items_2?limit=10").json()
        assert response == ITEMS[:10]

        schema = api.get_openapi_schema()["paths"]["/api/items_2"]["get"]
        # print(schema["parameters"])
        assert schema["parameters"] == [
            {
                "in": "query",
                "name": "someparam",
                "schema": {"title": "Someparam", "default": 0, "type": "integer"},
                "required": False,
            },
            {
                "in": "query",
                "name": "limit",
                "schema": {
                    "title": "Limit",
                    "default": 100,
                    "exclusiveMinimum": 0,
                    "type": "integer",
                },
                "required": False,
            },
            {
                "in": "query",
                "name": "offset",
                "schema": {
                    "title": "Offset",
                    "default": 0,
                    "exclusiveMinimum": -1,
                    "type": "integer",
                },
                "required": False,
            },
        ]

    def test_case3(self):
        response = client.get("/items_3?skip=5").json()
        assert response == ITEMS[5:10]

        schema = api.get_openapi_schema()["paths"]["/api/items_3"]["get"]
        # print(schema)
        assert schema["parameters"] == [
            {
                "in": "query",
                "name": "skip",
                "schema": {"title": "Skip", "type": "integer"},
                "required": True,
            }
        ]

    def test_case4(self):
        response = client.get("/items_4?page=2").json()
        assert response.get("results") == ITEMS[10:20]
        assert response.get("count") == 100
        assert response.get("next") == "http://testlocation/?page=3"
        assert response.get("previous") == "http://testlocation/"

        schema = api.get_openapi_schema()["paths"]["/api/items_4"]["get"]
        # print(schema)
        assert schema["parameters"] == [
            {
                "in": "query",
                "name": "page",
                "schema": {
                    "title": "Page",
                    "default": 1,
                    "exclusiveMinimum": 0,
                    "type": "integer",
                },
                "required": False,
            },
            {
                "in": "query",
                "name": "page_size",
                "schema": {
                    "title": "Page Size",
                    "default": 10,
                    "exclusiveMaximum": 200,
                    "type": "integer",
                },
                "required": False,
            },
        ]

    def test_case5(self):
        response = client.get("/items_5?page=2").json()
        assert response == ITEMS[10:20]

        schema = api.get_openapi_schema()["paths"]["/api/items_5"]["get"]
        # print(schema)
        assert schema["parameters"] == [
            {
                "in": "query",
                "name": "page",
                "schema": {
                    "title": "Page",
                    "default": 1,
                    "exclusiveMinimum": 0,
                    "type": "integer",
                },
                "required": False,
            }
        ]



from typing import Any, List, Union

from django.http.request import HttpRequest
from ninja.types import DictStrAny
from pydantic import BaseModel as PydanticModel, Field

from ninja_extra.types import PermissionType


class RouteContext(PydanticModel):
    class Config:
        arbitrary_types_allowed = True

    permission_classes: PermissionType = Field([])
    request: Union[Any, HttpRequest, None] = None
    args: List[Any] = Field([])
    kwargs: DictStrAny = Field({})



import threading
from typing import Any, Optional, cast

from django.apps import apps
from django.conf import Settings, settings
from django.dispatch import receiver
from injector import Binder, Module, singleton

from ninja_extra.controllers.route.context import RouteContext

from .signals import route_context_finished, route_context_started


@receiver([route_context_started, route_context_finished], sender=RouteContext)
def route_context_handler(
    *args: Any, route_context: Optional[RouteContext] = None, **kwargs: Any
) -> None:
    app = cast(Any, apps.get_app_config("ninja_extra"))
    app.ninja_extra_module.set_route_context(route_context)


class NinjaExtraModule(Module):
    def __init__(self) -> None:
        self._local = threading.local()

    def set_route_context(self, route_context: RouteContext) -> None:
        self._local.route_context = route_context

    def get_route_context(self) -> Optional[RouteContext]:
        try:
            return self._local.route_context  # type:ignore
        except AttributeError:
            return None

    def configure(self, binder: Binder) -> None:
        binder.bind(Settings, to=settings, scope=singleton)  # type: ignore
        binder.bind(RouteContext, to=lambda: self.get_route_context())  # type: ignore



import pytest

from ninja_extra.lazy import LazyStrImport


class TestLazyClassImport:
    """Test lazy import"""


def test_lazy_str_import_works():
    lazy_import = LazyStrImport("tests.test_lazy_import.TestLazyClassImport")
    instance = lazy_import()
    assert isinstance(instance, TestLazyClassImport)

    with pytest.raises(Exception) as exc:
        lazy_import = LazyStrImport("tests.test_lazy_import.InvalidLazyClassImport")
        lazy_import()



from django.contrib import admin
from django.urls import path

from ninja_extra import NinjaExtraAPI

from .controllers import EventController

api = NinjaExtraAPI()
api.register_controllers(EventController)


urlpatterns = [
    path("admin/", admin.site.urls),
    path("api/", api.urls),
]



import pytest
from django.contrib.auth.models import Group, Permission

from ninja_extra import status
from ninja_extra.exceptions import APIException, NotFound
from ninja_extra.shortcuts import get_object_or_exception, get_object_or_none


class CustomException(APIException):
    status_code = status.HTTP_400_BAD_REQUEST


class TestModelUtils:
    @pytest.mark.django_db
    def test_get_object_or_exception_should_return_object(self):
        group = Group.objects.create(name="_groupowner")
        saved_perm = get_object_or_exception(Group, id=group.id)
        assert saved_perm

    @pytest.mark.django_db
    def test_get_object_or_exception_should_raise_exception(self):
        with pytest.raises(NotFound) as exception_info:
            get_object_or_exception(Permission, id=0)
        assert "Permission with id = 0  was not found" == exception_info.value.message
        assert exception_info.value.status_code == 404

    @pytest.mark.django_db
    def test_get_object_or_exception_with_error_message(self):
        with pytest.raises(NotFound) as exception_info:
            get_object_or_exception(Permission, error_message="Not found", id=0)
        assert "Not found" in str(exception_info.value.message)

    @pytest.mark.django_db
    def test_get_object_or_exception_with_custom_exception(self):
        with pytest.raises(CustomException) as exception_info:
            get_object_or_exception(
                Permission, error_message="Bad Request", id=0, exception=CustomException
            )
        assert "Bad Request" in str(exception_info.value.message)
        assert exception_info.value.status_code == 400

    @pytest.mark.django_db
    def test_get_object_or_none_with_error_message(self):
        obj = get_object_or_none(Permission, id=0)
        assert obj is None



from typing import TYPE_CHECKING

from django.http import HttpRequest

from ninja_extra.permissions.base import SAFE_METHODS, BasePermission

if TYPE_CHECKING:
    from ninja_extra.controllers.base import APIController


class AllowAny(BasePermission):
    """
    Allow any access.
    This isn't strictly required, since you could use an empty
    permission_classes list, but it's useful because it makes the intention
    more explicit.
    """

    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        return True


class IsAuthenticated(BasePermission):
    """
    Allows access only to authenticated users.
    """

    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        user = request.user or request.auth  # type: ignore
        return bool(user and user.is_authenticated)


class IsAdminUser(BasePermission):
    """
    Allows access only to admin users.
    """

    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        user = request.user or request.auth  # type: ignore
        return bool(user and user.is_staff)  # type: ignore


class IsAuthenticatedOrReadOnly(BasePermission):
    """
    The request is authenticated as a user, or is a read-only request.
    """

    def has_permission(self, request: HttpRequest, controller: "APIController") -> bool:
        user = request.user or request.auth  # type: ignore
        return bool(request.method in SAFE_METHODS or user and user.is_authenticated)



from typing import Any, Optional, Type, cast, no_type_check

from django.db.models import Model, QuerySet
from ninja.types import DictStrAny, TCallable

from .exceptions import APIException, NotFound


@no_type_check
def fail_silently(func: TCallable, *args: Any, **kwargs: Any) -> Optional[Any]:
    try:
        return func(*args, **kwargs)
    except (Exception,):
        pass
    return None


@no_type_check
def _get_queryset(klass: Type[Model]) -> QuerySet:
    # If it is a model class or anything else with ._default_manager
    if hasattr(klass, "_default_manager"):
        return cast(QuerySet, klass._default_manager.all())
    return cast(QuerySet, klass)


@no_type_check
def get_object_or_exception(
    klass: Type[Model],
    error_message: str = None,
    exception: Type[APIException] = NotFound,
    **kwargs: Any
) -> Any:
    queryset = _get_queryset(klass)
    _validate_queryset(klass, queryset)
    try:
        return queryset.get(**kwargs)
    except queryset.model.DoesNotExist as ex:
        if error_message:
            message = error_message
        else:
            message = "{} with {} was not found".format(
                queryset.model._meta.object_name, _format_dict(kwargs)
            )
        raise exception(message=message) from ex


def _format_dict(table: DictStrAny) -> str:
    table_str = ""
    for k, v in table.items():
        table_str += "{} = {} ".format(str(k), str(v))
    return table_str


@no_type_check
def get_object_or_none(klass: Type[Model], **kwargs: Any) -> Optional[Any]:
    queryset = _get_queryset(klass)
    _validate_queryset(klass, queryset)
    try:
        return queryset.get(**kwargs)
    except (queryset.model.DoesNotExist, KeyError):
        return None


@no_type_check
def _validate_queryset(klass: Type[Model], queryset: Any) -> None:
    if not hasattr(queryset, "get"):
        klass__name = (
            klass.__name__ if isinstance(klass, type) else klass.__class__.__name__
        )
        raise ValueError(
            "First argument must be a Model, Manager, "
            "or QuerySet, not '%s'." % klass__name
        )



from typing import Any, Type, Union
from unittest.mock import Mock

import pytest
from ninja import Schema

from ninja_extra import APIController, NinjaExtraAPI, route, router
from ninja_extra.controllers import RouteFunction
from ninja_extra.controllers.base import MissingRouterDecoratorException
from ninja_extra.controllers.response import ControllerResponse, Detail, Id, Ok
from ninja_extra.controllers.router import ControllerRouter
from ninja_extra.permissions.common import AllowAny


class SomeController(APIController):
    pass


class SomeControllerWithInject(APIController):
    def __init__(self, a: str):
        pass


class SomeControllerWithRoute(APIController):
    @route.get("/example")
    def example(self):
        pass

    @route.get("/example/{ex_id}")
    def example2(self, ex_id: str):
        pass


@router("", tags=["new tag"])
class SomeControllerWithRouter(APIController):
    auto_import = False  # disable auto_import of the controller

    @route.get("/example")
    def example(self):
        pass

    @route.get("/example/{ex_id}")
    def example2(self, ex_id: str):
        return self.create_response(ex_id, status_code=302)

    @route.get("/example/{ex_id}/ok")
    def example_with_ok_response(self, ex_id: str):
        return self.Ok(ex_id)

    @route.get("/example/{ex_id}/id")
    def example_with_id_response(self, ex_id: str):
        return self.Id(ex_id)


class TestAPIController:
    def test_controller_should_have_preset_properties(self):
        api = NinjaExtraAPI()
        assert SomeController.tags == ["some"]
        assert SomeController._path_operations == {}
        assert SomeController.permission_classes == [AllowAny]
        assert SomeController._router is None
        assert SomeController.api is None
        assert SomeController.registered is False

        with pytest.raises(MissingRouterDecoratorException) as ex:
            api.register_controllers(SomeController)
        assert "Controller Router not found" in str(ex.value)

    def test_controller_with_router_should_have_preset_properties(self):
        api = NinjaExtraAPI()
        assert SomeControllerWithRouter.permission_classes == [AllowAny]
        assert isinstance(SomeControllerWithRouter._router, ControllerRouter)
        assert SomeControllerWithRouter.api is None
        assert SomeControllerWithRouter.registered is False

        api.register_controllers(SomeControllerWithRouter)
        assert SomeControllerWithRouter.api == api
        assert SomeControllerWithRouter.registered

    def test_controller_should_wrap_with_inject(self):
        assert not hasattr(SomeController.__init__, "__bindings__")
        assert hasattr(SomeControllerWithInject.__init__, "__bindings__")

    def test_controller_should_have_path_operation_list(self):
        assert len(SomeControllerWithRoute._path_operations) == 2

        route_function: RouteFunction = SomeControllerWithRoute.example
        path_view = SomeControllerWithRoute._path_operations.get(str(route_function))
        assert path_view, "route doesn't exist in controller"
        assert len(path_view.operations) == 1

        operation = path_view.operations[0]
        assert operation.methods == route_function.route.route_params.methods
        assert operation.operation_id == route_function.route.route_params.operation_id

    def test_controller_route_definition_should_return_instance_route_definitions(self):
        assert len(SomeControllerWithRoute._path_operations) == 2
        for route_definition in SomeControllerWithRoute.get_route_functions():
            assert isinstance(route_definition, RouteFunction)


class TestAPIControllerResponse:
    class CustomResponse(ControllerResponse):
        status_code = 502

        def __init__(self, email: str, name: str):
            super().__init__()
            self.email = email
            self.name = name

        class CustomSchema(Schema):
            email: str
            name: str

        @classmethod
        def get_schema(cls) -> Union[Schema, Type[Schema], Any]:
            return cls.CustomSchema

        def convert_to_schema(self) -> Any:
            return self.CustomSchema.from_orm(self)

    ok_response = Ok("OK")
    id_response = Id("ID")
    detail_response = Detail(dict(errors=[dict(test="passed")]), status_code=302)

    def test_controller_response(self):
        # OK Response
        assert self.ok_response.get_schema() == Ok.get_schema()
        assert self.ok_response.convert_to_schema() == Ok.get_schema()(message="OK")
        assert self.ok_response.status_code == Ok.status_code
        # ID Response
        assert self.id_response.get_schema() == Id.get_schema()
        assert self.id_response.convert_to_schema() == Id.get_schema()(id="ID")
        assert self.id_response.status_code == Id.status_code
        # Detail Response
        assert self.detail_response.get_schema() == Detail.get_schema()
        assert self.detail_response.convert_to_schema() == Detail.get_schema()(
            message=dict(errors=[dict(test="passed")])
        )
        assert self.id_response.status_code != Detail.status_code
        # CustomResponse Response
        custom_response = self.CustomResponse(email="some_email", name="some_name")
        assert custom_response.get_schema() == self.CustomResponse.get_schema()
        assert custom_response.convert_to_schema() == self.CustomResponse.get_schema()(
            email="some_email", name="some_name"
        )
        assert custom_response.status_code == self.CustomResponse.status_code

    def test_controller_response_works(self):
        detail = Detail("5242", status_code=302)
        result = SomeControllerWithRouter.example2(request=Mock(), ex_id="5242")
        assert isinstance(result, tuple)
        assert result[1] == detail.convert_to_schema()
        assert result[0] == detail.status_code

        ok_response = Ok("5242")
        result = SomeControllerWithRouter.example_with_ok_response(
            request=Mock(), ex_id="5242"
        )
        assert isinstance(result, tuple)
        assert result[1] == ok_response.convert_to_schema()
        assert result[0] == ok_response.status_code

        id_response = Id("5242")
        result = SomeControllerWithRouter.example_with_id_response(
            request=Mock(), ex_id="5242"
        )
        assert isinstance(result, tuple)
        assert result[1] == id_response.convert_to_schema()
        assert result[0] == id_response.status_code
